#include <click/config.h>
#include "seriallink2.hh"
#include <click/error.hh>
#include <click/confparse.hh>
#include <click/glue.hh>
#include <click/straccum.hh>
CLICK_DECLS

/*
 * seriallink.cc,hh
 * Eric Freudenthal (freudenthal@nyu.edu) & Tao Zhao (taozhao@cs.nyu.edu)
 * and Eddie Kohler
 *
 * Copyright (c) 2002 New York University
 * Copyright (c) 2002 International Computer Science Institute
 *
 * For integration into the click modular router (MIT).
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, subject to the conditions
 * listed in the Click LICENSE file. These conditions include: you must
 * preserve this copyright notice, and you cannot mention the copyright
 * holders in advertising related to the Software without their permission.
 * The Software is provided WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED. This
 * notice is a summary of the Click LICENSE file; the license in that file is
 * legally binding.
 */

SerialLink::SerialLink()
  : Element(1, 1), _qhead(0), _qtail(0), _size(0), _timer(this),
    _nreceived(0), _ndrops(0)
{
  MOD_INC_USE_COUNT;
}

SerialLink::~SerialLink()
{
  MOD_DEC_USE_COUNT;
}

int
SerialLink::configure(Vector<String> &conf, ErrorHandler *errh)
{
  _capacity = 1000;
  return cp_va_parse(conf, this, errh,
		     cpInterval, "delay (us)", &_link_latency,
		     cpUnsigned, "bandwidth (Kbaud)", &_bandwidth,
		     cpUnsigned, "queue capacity (messages)", &_capacity,
		     0);
}

int
SerialLink::initialize(ErrorHandler *)
{
  _timer.initialize(this);
  return 0;
}

void
SerialLink::cleanup(CleanupStage)
{
  while (_qhead) {
    Packet *q = _qhead;
    _qhead = _qhead->next();
    q->kill();
  }
}

struct timeval
SerialLink::serialization_delay(const Packet *p) const
{
  uint32_t p_bits = p->length() * 8;
  timeval delay;
  delay.tv_sec = 0;
  delay.tv_usec = (p_bits * 1000) / _bandwidth;
  while (delay.tv_usec >= 1000000)
    delay.tv_sec++, delay.tv_usec -= 1000000;
  return delay;
}

void
SerialLink::push(int, Packet *p)
{
  _nreceived++;
  
  if (_size >= _capacity) {	// check _size; might actually have room
    int real_size = _size;
    
    // t = now + _link_latency, so any packet scheduled for expulsion at or
    // after t has completely entered the channel by now, and need not be
    // considered part of 'real_size'.
    struct timeval t;
    click_gettimeofday(&t);
    t += _link_latency;
    
    for (Packet *q = _qhead; q && q->timestamp_anno() >= t; q = q->next())
      real_size--;
    
    if (real_size >= _capacity) {
      p->kill();
      _ndrops++;
      return;
    }
  }
  
  if (_size == 0) {
    click_gettimeofday(&p->timestamp_anno());
    p->timestamp_anno() += serialization_delay(p) + _link_latency;
    _timer.schedule_at(p->timestamp_anno());
    _qhead = _qtail = p;
    p->set_next(0);
    _size = 1;
  } else {
    p->timestamp_anno() = _qtail->timestamp_anno() + serialization_delay(p);
    _qtail->set_next(p);
    _qtail = p;
    p->set_next(0);
    _size++;
  }
}

void
SerialLink::run_timer()
{
  timeval now;
  click_gettimeofday(&now);
  while (_qhead && _qhead->timestamp_anno() <= now) {
    Packet *q = _qhead;
    _qhead = q->next();
    _size--;
    q->set_next(0);
    output(0).push(q);
  }
  if (!_qhead)
    _qtail = 0;
  else
    _timer.reschedule_at(_qhead->timestamp_anno());
}


int
SerialLink::write_handler(const String &arg, Element *e,
			  void *thunk, ErrorHandler *errh)
{
  SerialLink *u = (SerialLink *)e;
  switch ((intptr_t)thunk) {
   case 1: {
     uint32_t bandwidth;
     if (!cp_unsigned(cp_uncomment(arg), &bandwidth) || bandwidth < 1)
       return errh->error("bandwidth must be unsigned integer >= 1");
     u->_bandwidth = bandwidth;
     break;
   }
   default:
    return errh->error("<internal>");
  }
  return 0;
}



String
SerialLink::read_param(Element *e, void *vparam)
{
  SerialLink *u = (SerialLink *)e;

  switch ((int)vparam) {
   case 0:                      // latency
     return cp_unparse_interval(u->_link_latency) + "\n";
   case 1:                      //  bandwidth
     return String(u->_bandwidth) + "\n";
   case 2:                      // queue capacity
     return String(u->_capacity) + "\n";
   case 3:                      // # items enqueued
     return String(u->_size) + "\n";
   case 4:                      // # packets dropped
     return String(u->_ndrops) + "\n";
   case 5: {			// callback scheduler: # events scheduled
     StringAccum sa;
     if (u->_size > 0) {
       timeval now;
       click_gettimeofday(&now);
       sa << "now=" << now << ", head=" << u->_qhead->timestamp_anno()
	  << (now > u->_qhead->timestamp_anno() ? " - outdated\n" : " - current\n");
     } else
       sa << "empty\n";
     return sa.take_string();
   }
   case 6:
    return String(u->_nreceived) + "\n";
  }

  return String("SerialLink::read_param: bad discriminator\n");
}

void
SerialLink::add_handlers()
{
  add_read_handler("delay", read_param, (void *)0);
  add_read_handler("bandwidth", read_param, (void *)1);
  add_write_handler("bandwidth", write_handler, (void *)1);
  add_read_handler("max_queue", read_param, (void *)2);
  add_read_handler("qlen", read_param, (void *)3);
  add_read_handler("drops", read_param, (void *)4);
  add_read_handler("time", read_param, (void *)5);
  add_read_handler("packetRecCount", read_param, (void *)6);
}

CLICK_ENDDECLS
EXPORT_ELEMENT(SerialLink)
